from contextlib import asynccontextmanager
from hashlib import md5
from typing import AsyncGenerator

from fastapi import FastAPI, Request, Response, status
from fastapi.exception_handlers import http_exception_handler, request_validation_exception_handler
from fastapi.exceptions import RequestValidationError, StarletteHTTPException
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.openapi.docs import get_swagger_ui_html
from fastapi.responses import HTMLResponse, JSONResponse
from fastapi.routing import APIRoute
from httpx import AsyncClient
from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.trace import Status, StatusCode

from app.api.api import api_router
from app.api.miscellaneous_endpoints import miscellaneous_router
from app.core.config import settings

description = """
This is the s3 proxy service from the CloWM Service.
"""


def custom_generate_unique_id(route: APIRoute) -> str:
    return f"{route.tags[-1]}-{route.name}"


@asynccontextmanager
async def lifespan(fastapi_app: FastAPI) -> AsyncGenerator[None, None]:  # pragma: no cover
    # Create a http client once instead for every request and attach it to the app
    async with AsyncClient() as client:
        fastapi_app.requests_client = client  # type: ignore[attr-defined]
        yield


app = FastAPI(
    title="CloWM S3-Proxy Service",
    version="2.0.0",
    description=description,
    contact={
        "name": "Daniel Goebel",
        "url": "https://ekvv.uni-bielefeld.de/pers_publ/publ/PersonDetail.jsp?personId=223066601",
        "email": "dgoebel@techfak.uni-bielefeld.de",
    },
    generate_unique_id_function=custom_generate_unique_id,
    # license_info={"name": "MIT", "url": "https://mit-license.org/"},
    root_path=settings.API_PREFIX,
    openapi_url=None,  # create it manually to enable caching on client side
    lifespan=lifespan,
)
if settings.API_PREFIX:  # pragma: no cover
    app.servers.insert(0, {"url": settings.API_PREFIX})

if settings.OTLP_GRPC_ENDPOINT is not None and len(settings.OTLP_GRPC_ENDPOINT) > 0:  # pragma: no cover
    resource = Resource(attributes={SERVICE_NAME: "clowm-s3proxy-service"})
    provider = TracerProvider(resource=resource)
    provider.add_span_processor(
        BatchSpanProcessor(OTLPSpanExporter(endpoint=settings.OTLP_GRPC_ENDPOINT, insecure=True))
    )
    trace.set_tracer_provider(provider)

    @app.exception_handler(StarletteHTTPException)
    async def trace_http_exception_handler(request: Request, exc: StarletteHTTPException) -> Response:
        current_span = trace.get_current_span()
        current_span.set_status(Status(StatusCode.ERROR))
        current_span.record_exception(exc)
        return await http_exception_handler(request, exc)

    @app.exception_handler(RequestValidationError)
    async def trace_validation_exception_handler(request: Request, exc: RequestValidationError) -> JSONResponse:
        current_span = trace.get_current_span()
        current_span.set_status(Status(StatusCode.ERROR))
        current_span.record_exception(exc)
        return await request_validation_exception_handler(request, exc)


FastAPIInstrumentor.instrument_app(
    app, excluded_urls="health,docs,openapi.json", tracer_provider=trace.get_tracer_provider()
)

# Enable gzip compression for large responses
app.add_middleware(GZipMiddleware, minimum_size=500)

# Include all routes
app.include_router(api_router)
app.include_router(miscellaneous_router)


# manually add Swagger UI route
async def swagger_ui_html(req: Request) -> HTMLResponse:
    return get_swagger_ui_html(
        openapi_url=app.root_path + "/openapi.json",
        title=app.title + " - Swagger UI",
        swagger_favicon_url="/favicon.ico",
    )


app.add_route("/docs", swagger_ui_html, include_in_schema=False)

# Hash openapi.json content and ensure the serialization is the same as the one in the route
openapi_hash = md5(JSONResponse(app.openapi()).body).hexdigest()


# Create Custom route for OpenAPI schema
async def openapi(req: Request) -> Response:
    # If schema on clients side is still valid, return empty Body with 304 response code (client will use cached body)
    if req.headers.get("If-None-Match") == openapi_hash:
        return Response(status_code=status.HTTP_304_NOT_MODIFIED)
    # Return openapi.json with ETag header
    return JSONResponse(app.openapi(), headers={"ETag": openapi_hash})


app.add_route("/openapi.json", openapi, include_in_schema=False)
